#library "DJBACS"
#include "zcommon.acs"
#nocompact

/*###########################################################################*/
/* CONSTANTS ################################################################*/
/*###########################################################################*/

/* ****************************************************************
// CORE CONSTANTS
// ***************************************************************/
  /*
   * Tics per second.
   *
   * The number of tics for 1 second of time.
   *
   * Note: Never change this.
   */
  #define TICS_PER_SEC 35

  /*
   * Max Client Count
   *
   * The maximum number of players connected to the server.
   */
  #define MAX_CLIENT_COUNT 64

  #define MAX_WEAPONS 13
  #define MAX_AMMOS 4

  /*
   * Restart Countdown Max Time.
   *
   * The maximum number of seconds the countdown will be for restarting the
   * game when all players have left the game.
   */
  #define RESTART_COUNTDOWN_MAXTIME 60

  /*
   * Excited Laugh Death Count.
   *
   * The least amount of player deaths (within one second) required for an
   * excited laugh to play instead of the standard laugh.
   *
   * For example, if more than 1 player(s) die within one second, then hell
   * will laugh with excitement.
   */
  #define EXCITED_LAUGH_DEATH_COUNT 1

  /*
   * Elite monster minimum health.
   *
   * The minimum health for an elite monster, when health is being scaled.
   *
   * Tip:
   * This is also used to determine XP rewards/drops from the death of
   * a monster.
   */
  #define ELITE_MIN_HEALTH 10000

  /*
   * Script #s
   */
  #define SCR_MAP 1
  #define SCR_OPEN 10
  #define SCR_OPEN_2 11
  #define SCR_OPEN_3 12
  #define SCR_ENTER 20
  #define SCR_ENTER_2 21
  #define SCR_RETURN 30
  #define SCR_RETURN_2 31
  #define SCR_DEATH 40
  #define SCR_DEATH_2 49
  #define SCR_RESPAWN 50
  #define SCR_LIGHTNING 60
  #define SCR_UNLOADING 70
  #define SCR_DISCONNECT 80

  #define SCR_NO_PLAYERS 90

  //#define USE_HITSCAN_CVAR "coopbullets_cl_use_hitscan"

  #define SCR_UNIQUE_TID 114
  #define SCR_MAX_LEVEL 125
  #define SCR_SHOT_SPREAD 116
  #define SCR_MON_SPAWN 766
  #define SCR_MON_DEATH 206
  #define SCR_LEVEL_PLAYER 207
  #define SCR_LEVEL_PLAYER_FX 208
  #define SCR_BACKPACK_RESPAWN 423

  //#define CVAR_XP_MAX_LVL "ad_xp_max_level"

/* ****************************************************************
// MONSTER CONSTANTS
// ***************************************************************/
  #define MON_CLASS_MAX_COUNT 24
  #define MON_TID_RANGE 10000
  #define MAP_MON_CLASS_MAX_LIMIT 60 // 60

/* ****************************************************************
// BACKPACK SPAWNER CONSTANTS
// ***************************************************************/
  /* Backpack Spawners Enabled
   *
   * Enable or disable the backpack spawners feature.
   */
  #define IS_BACKPACK_SPAWNERS_ENABLED 1

  /* Backpack Respawn Delay
   *
   * The respawn time (in seconds) for a backpack.
   */
  #define BP_RESPAWN_DELAY 30 // Use 30.

  /* Backpack classname.
   *
   * The system name of the backpack item to spawn.
   */
  #define BP_CLASSNAME "Backpack"

  /* Max Backpack Spawners.
   *
   * The maximum number of backpack spawners to create.
   */
  #define MAX_BP_SPAWNERS 10

/* ****************************************************************
// XP CONSTANTS
// ***************************************************************/
  /*
   * XP Classname.
   *
   * The system name of a point of XP.
   */
  #define XP_CLASSNAME "djbPoint"

  /*
   * Level Classname.
   *
   * The system name of a player level.
   */
  #define XP_LVL_CLASSNAME "djbLevel"
  #define XP_MAX_LVL_CLASSNAME "djbMaxLevel"

  /*
   * Points percent by monster health.
   *
   * E.g. Monster of 200HP will yield 10% (20 points) of it's health as points.
   */
  #define XP_GIVEAWAY_PERCENT_BY_MON_HP 10

  /*
   * Default XP Amount
   *
   * The default amount of points to give the player for a monster kill, if no
   * points setting can be found for that particular monster.
   */
  #define XP_GIVEAWAY_AMT_DEFAULT 10

  /*
   * XP Giveaway Distance Extension
   *
   * The extended distance for the range that players must be within to receive
   * a share of the points.
   */
  #define XP_GIVEAWAY_DISTANCE_EXT 512

  /*
   * Random Points Slice Minimum Amount
   *
   * The minimum amount of points a slice can be.  Set this to zero to be less
   * generous about distributing points to all players within range.
   */
  #define XP_RAND_PTS_SLICE_AMT_MIN 1

  /*
   * Random Points Slice Percentage
   *
   * The percentage amount to sliced (removed from, or added to) the amount of
   * points to be given away.
   */
  #define XP_RAND_PTS_SLICE_PERCENT 5

  /*
   * Level Max XP Increment Amount.
   *
   * The amount to increase max xp per level.
   * E.g. Level 2 = 100, Level 3 = 150, etc.
   */
  //#define XP_LVL_MAX_XP_INCR_AMT 1000
  //#define XP_LVL_MAX_XP_INCR_MAX_AMT 10000

  /*
   * Level Damage Factor (Percent)
   *
   * The % amount of damage factor adjusted per level-up.
   */
  #define XP_LVL_DMG_FACTOR 5 // Percent.

  /*
   * Max Level Damage Factor (Percent)
   *
   * The maximum % amount of damage factor adjusted to.
   */
  #define XP_MAX_LVL_DMG_FACTOR 75 // Percent.

  /*
   * Top Players Max Count
   *
   * The maximum number of top players to display on screen/scoreboard.
   */
  #define XP_TOP_PLAYERS_MAX_COUNT 5
  #define XP_TOP_PLAYERS_HUD_DELAY 30 // Use 30.

/* ****************************************************************
// CONFIG CONSTANTS
// ***************************************************************/
  //#define XP_MAX_REWARDS 7

  // Don't change these:
  #define IDX_CFG_MAX_COUNT 1000

  // Setting Type (1 = reward, 2 = weapon, 3 = ammo, 4 = other)
  #define IDX_CFG_TYPE_REWARD 1
  #define IDX_CFG_TYPE_WEAPON 2
  #define IDX_CFG_TYPE_AMMO 3
  #define IDX_CFG_TYPE_XP 9

  // Type Category (1 = weapon, 2 = ammo)
  //#define IDX_CFG_ITEM_WEAPON 1
  //#define IDX_CFG_ITEM_AMMO 2
  //#define IDX_CFG_ITEM_MAX 9

  // Weapon(Category) Setting Type (1 = name, 2 = amount)
  #define IDX_CFG_ITEM_WEAPON_CLASS 1 // Weapon class name.
  #define IDX_CFG_ITEM_WEAPON_AMT 2 // Weapon amount.

  // Ammo(Category) Setting Type (1 = name, 2 = amount)
  #define IDX_CFG_ITEM_AMMO_CLASS 1 // Ammo class name.
  #define IDX_CFG_ITEM_AMMO_AMT 2 // Ammo amount.

  // Reward(Category) Setting Type (1 = weapon, 2 = ammo)
  #define IDX_CFG_REWARD_WNUM 1 // Weapon number.
  #define IDX_CFG_REWARD_ANUM 2 // Ammo number.

  // XP
  #define IDX_CFG_XP_MAX_INCR 1 // Weapon amount.
  #define IDX_CFG_XP_MAX_INCR_LIMIT 2 // Weapon amount.

/* ****************************************************************
// REGENERATION CONSTANTS
// ***************************************************************/
  /* Backpack Spawners Enabled
   *
   * Enable or disable the backpack spawners feature.
   */
  #define IS_REGEN_ENABLED 0

  #define REGEN_MAX_TYPES 1
  #define REGEN_TYPE_AMMO "ammo"
  #define REGEN_TYPE_INV "inventory"
  str regen_attr[1][6] = {
    {"Stamina", REGEN_TYPE_AMMO, "StaminaRegenRate", REGEN_TYPE_AMMO, "StaminaRegenAmt", REGEN_TYPE_AMMO}
  };

/* ****************************************************************
// HUD
// ***************************************************************/
  // HUD: No players.
    #define HUD_NO_PLAYERS_TEXT "NO_PLAYERS"
    #define HUD_NO_PLAYERS_TYPE 0
    #define HUD_NO_PLAYERS_CHAN 500
    #define HUD_NO_PLAYERS_COLOR CR_WHITE
    #define HUD_NO_PLAYERS_X 320.0
    #define HUD_NO_PLAYERS_Y 40.0
    #define HUD_NO_PLAYERS_DURATION 0

  // HUD: "Team XP:" (Scoreboard)
    #define HUD_PLAYERS_SCORE_TEXT "TOP_PLAYERS"
    #define HUD_PLAYERS_SCORE_TYPE HUDMSG_FADEOUT
    #define HUD_PLAYERS_SCORE_CHAN 551
    #define HUD_PLAYERS_SCORE_COLOR CR_RED
    #define HUD_PLAYERS_SCORE_X 1.1
    #define HUD_PLAYERS_SCORE_Y 310.2
    #define HUD_PLAYERS_SCORE_DURATION 7.0
    #define HUD_PLAYERS_SCORE_FADETIME 3.0

  // HUD: [LEVEL UP BAR]
    #define HUD_LUP_FRAME_FONT "LVLUP"
    #define HUD_LUP_FRAME_TEXT "a"
    #define HUD_LUP_FRAME_TYPE HUDMSG_FADEINOUT
    #define HUD_LUP_FRAME_CHAN 600 // * + p
    #define HUD_LUP_FRAME_COLOR CR_UNTRANSLATED
    #define HUD_LUP_FRAME_X 320.0
    #define HUD_LUP_FRAME_Y 125.0
    #define HUD_LUP_FRAME_DURATION 7.5
    #define HUD_LUP_FRAME_FADETIME 2.5
    #define HUD_LUP_FRAME_FADEOUTTIME 2.5

    #define HUD_THEIR_LVLUP_FRAME_FONT "CPLVLUP"
    #define HUD_THEIR_LVLUP_FRAME_CHAN 900
    #define HUD_THEIR_LVLUP_FRAME_Y 344.0

  // HUD: [LEVEL UP BAR TEXT] "You've reached"
    #define HUD_LUP_TXT_REACHED_FONT "BigFont"
    #define HUD_LUP_TXT_REACHED_Y 120.0
    #define HUD_LUP_TXT_REACHED_CHAN 700 // * + p
    #define HUD_LUP_TXT_REACHED_COLOR CR_WHITE
    #define HUD_LUP_TXT_WORD_YOU "PLAYER_YOU" // LANGUAGE
    #define HUD_LUP_TXT_WORD_REACHED "PLAYER_REACHED" // LANGUAGE

    #define HUD_THEIR_LUP_TXT_REACHED_FONT "NORMAL"
    #define HUD_THEIR_LUP_TXT_REACHED_CHAN 910
    #define HUD_THEIR_LUP_TXT_REACHED_Y 340.0

  // HUD: [LEVEL UP BAR TEXT] "Level #"
    #define HUD_LUP_TXT_LEVEL_FONT "NORMAL"
    #define HUD_LUP_TXT_LEVEL_COLOR CR_RED
    #define HUD_LUP_TXT_LEVEL_CHAN 800 // * + p
    #define HUD_LUP_TXT_LEVEL_Y 134.0
    #define HUD_LUP_TXT_WORD_LEVEL "PLAYER_LEVEL"

    #define HUD_THEIR_LUP_TXT_LEVEL_COLOR CR_WHITE
    #define HUD_THEIR_LUP_TXT_LEVEL_CHAN 920
    #define HUD_THEIR_LUP_TXT_LEVEL_Y 350.0
    #define MAX_CLASSES 1 
    #define MAX_REWARDS 127

/*###########################################################################*/
/* VARIABLES ################################################################*/
/*###########################################################################*/
  global int 25:djb[]; // Do not change this.

  /* ****************************************************************
  // FACTORY VARIABLES
  // ***************************************************************/
    // Indexes for map settings.
    int xa = 100; // Do not change this.
    int xb = 101; // Do not change this.
    int xc = 102; // Do not change this.
    int xd = 103; // Do not change this.

  /* ****************************************************************
  // CORE VARIABLES
  // ***************************************************************/
    /*
     * Gate Thrust Directions. 0 = Forward, 1 = Right, 2 = Backward, 3 = Left
     *                         4 = Forward/Right, 5 = Backward/Right, 6 = Backward/Left, 7 = Forward/Left
     */
    // int gate_thrust_dir[8] = { 0, 64, 128, 192, 32, 96, 160, 224 };
    // Reference: 64 = North, 128 = West, 192 = South, 0 = East?
    //            32 = North/East, 224 = South/East, 160 = South/West, 96 = North/West
    //int gate_thrust_dir[8] = { 64, 0, 192, 128, 32, 224, 160, 96 };

  /* ****************************************************************
  // PLAYER VARIABLES
  // ***************************************************************/
    /* Player TIDs.
     *
     * Keep track of each player's TID.
     */
    int djb_player_tids[MAX_CLIENT_COUNT]; // Maximum 64 players.

    /* Player death tracking.
     *
     * Keep track of each player's death count.
     */
    int djb_player_deaths[MAX_CLIENT_COUNT];
    int djb_recent_p_death_count;

    /* Player alive count.
     *
     * Keep track of # of players alive.  The count should only be
     * adjusted at times when a player is dead (has not respawned yet), or when a
     * player respawns..
     *
     * E.g. A player can be in jail and still be considered as alive.
     */
    int djb_player_alive_count = 0;

    /* Player alive tracking.
     *
     * Keep track of each player that is alive.
     *
     * Handled in events: ENTER, RESPAWN, DEATH, and DISCONNECT.
     */
    int djb_player_alive[MAX_CLIENT_COUNT];

    // Player's points array.
    // Note:
    // This is used for looping purposes only.  It does not (and never does)
    // contain the correct/current amount of points for any player.  It is
    // constantly reset, and used for loops.
    int djb_player_points[MAX_CLIENT_COUNT];
    int djb_player_levels[MAX_CLIENT_COUNT];

  /* ****************************************************************
  // TEAM VARIABLES
  // ***************************************************************/
    /* Team names.
     *
     * An array of team names.
     *
     * Tip: Replace words with LANGUAGE vars.
     */
    str djb_team_name[8] = {"Blue", "Red", "Green", "Yellow", "Black", "White", "Orange", "Purple"};
    
    /* Team colors.
     *
     * An array of team colors.
     *
     * Blue, Red, Green, Yellow, Ddjb Gray, Light Gray, Orange, Purple.
     */
    str djb_team_color[8] = {"\ch", "\cg", "\cd", "\ck", "\cu", "\cc", "\ci", "\ct"};

  /* ****************************************************************
  // BACKPACK VARIABLES
  // ***************************************************************/
    /* Backpack Spawner TIDs.
     *
     * Keep track of each backpack spawner's TID.
     */
    int djb_bp_spawner_tids[MAX_BP_SPAWNERS];

    /* Backpack TIDs.
     *
     * Keep track of each backpack spawner's TID.
     */
    int djb_bp_tids[MAX_BP_SPAWNERS];

    /* Backpacks Spawned.
     *
     * Keep track of when backpacks are spawned.
     */
    bool djb_bps_spawned;

  /* ****************************************************************
  // XP VARIABLES
  // ***************************************************************/
    // Max # of points to give player, if player kills monster with higher health
    // than the highest possible (in mon_hp_to_xp[]).
    //int mon_hp_to_xp_max = 10000;

    // Top players array.
    int djb_top_players[XP_TOP_PLAYERS_MAX_COUNT];

    // Max level for player.
    int djb_max_level;

    //int djb_max_xp_incr_amt;
    //int djb_max_xp_incr_amt_limit;

    // Total # of points for "Team XP:"
    // Note:
    // This is used for looping purposes, and therefore is constantly reset and
    // re-used.
    int djb_cur_total_points;
    int djb_cur_total_levels;

  /* ****************************************************************
  // MONSTER VARIABLES
  // ***************************************************************/
    str ad_mon_class[MON_CLASS_MAX_COUNT] = {
      "Zombieman2",
      "ShotgunGuy2",
      "ChaingunGuy2",
      "DoomImp2",
      "Demon2",
      "BaronOfHell2",
      "HellKnight2",
      "Cacodemon2",
      "LostSoul2",
      "PainElemental2",
      "Revenant2",
      "Fatso2",
      "Archvile2",
      "Cyberdemon2",
      "Arachnotron2",
      "SpiderMastermind2",
      "WolfensteinSS2",
      "Abaddon2",
      "Belphegor2",
      "BloodDemon2",
      "Cacolantern2",
      "DarkImp2",
      "SuperShotgunGuy2"
    };

    int ad_mon_base_tid[MON_CLASS_MAX_COUNT] = {
      100000,
      200000,
      300000,
      400000,
      500000,
      600000,
      700000,
      800000,
      900000,
      1000000,
      1100000,
      1200000,
      1300000,
      1400000,
      1500000,
      1600000,
      1700000,
      1800000,
      1900000,
      2000000,
      2100000,
      2200000,
      2300000,
	  2400000
    };

    int ad_mon_tid[MON_CLASS_MAX_COUNT] = {
      100000,
      200000,
      300000,
      400000,
      500000,
      600000,
      700000,
      800000,
      900000,
      1000000,
      1100000,
      1200000,
      1300000,
      1400000,
      1500000,
      1600000,
      1700000,
      1800000,
      1900000,
      2000000,
      2100000,
      2200000,
      2300000,
	  2400000
    };

    int ad_mon_count_total;
    int ad_mon_count[MON_CLASS_MAX_COUNT] = {
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
	  0
    };

    int ad_mon_hp[MON_CLASS_MAX_COUNT] = {
      20, // ZombieMan
      30, // ShotgunGuy
      70, // ChaingunGuy
      60, // DoomImp
      150, // Demon
      1000, // BaronOfHell
      500, // HellKnight
      400, // CacoDemon
      100, // LostSoul
      400, // PainElemental
      300, // Revenant
      600, // Fatso
      700, // Archvile
      4000, // Cyberdemon
      500, // Arachnotron
      3000, // SpiderMasterMind
      50, // WolfensteinSS
      1200, // Abaddon 
      1500, // Belphegor
      300, // BloodDemon
      800, // Cacolantern
      120, // DarkImp
      1200, // Hectebus
      120 // SuperShotgunGuy
    };

    int ad_map_mon_max_limit;
    int ad_remove_corpse_delay;



/*###########################################################################*/
/* SCRIPTS ##################################################################*/
/*###########################################################################*/
  /* **************************************************************************
  // OPEN
  // *************************************************************************/
    /*
     * The level runs an OPEN script when it is started up. The purpose of OPEN
     * scripts is to run the overall level. You can guarantee that it will always
     * run, and only run once. They often contain code which monitors the level and
     * waits for certain circumstances.
     *
     * In Survival mode, only OPEN CLIENSIDE scripts are executed after connecting
     * to the server. Non-clientside OPEN scripts will not execute.
     * NOTE: Non-clientside OPEN scripts will execute, however, when (if you
     * connected during a warmup time) the warmup ends(fight begins).
     *
     * TIP: OPEN CLIENTSIDE scripts will always occur after OPEN scripts.
     */
    script SCR_OPEN OPEN {
      //InitializeMap();

      // If Gamemode is NOT deathmatch.
      if (!IsGameTypeDeathmatch()) {
        SetActivator(0);

        if (IsAnyPlayerJoined()) {
          ClearNoPlayersMessageHUD();

          Game();
        }
        else {
          DisplayNoPlayersMessageHUD();
        }

        Delay(10);

        Restart;
      }
    }
	  script 998 ENTER
	  {  
   IncreaseMaximumHealth(0);
     }
    /* ****************************************************************
    // OPEN - HUD: TOP PLAYERS
    // ***************************************************************/
      script SCR_OPEN_2 OPEN {
        // If Gamemode is NOT deathmatch.
        if (!IsGameTypeDeathmatch() && !IsGameTypeTitle()) {
          // Set HUD Properties.
          str hud_message = StrParam(l:HUD_PLAYERS_SCORE_TEXT, s:"");

          int p, c, t;
          int cur_count = 0;
          int cur_max_points;
          int cur_max_player;
          int cur_player_points;
          int cur_player_level;
          int cur_player_num_one_based;
          djb_cur_total_points = 0;
          djb_cur_total_levels = 0;

          // Reset
          for (p = 0; p < XP_TOP_PLAYERS_MAX_COUNT; p++) {
            djb_top_players[p] = 0;
          }

          // Loop through players.
          for (p = 0; p < MAX_CLIENT_COUNT; p++) {
            djb_player_points[p] = 0; // Reset
            djb_player_levels[p] = 0; // Reset

            if (IsPlayerEligibleForPoints(p)) {
              cur_count++;
              cur_player_points = CheckActorInventory(GetPlayerTIDByNum(p), XP_CLASSNAME);
              cur_player_level = CheckActorInventory(GetPlayerTIDByNum(p), XP_LVL_CLASSNAME);
              djb_cur_total_points += cur_player_points;
              djb_cur_total_levels += cur_player_level;
              cur_player_num_one_based = p+1;
              djb_player_points[p] = cur_player_points;
              djb_player_levels[p] = cur_player_level;

              for (t = 0; t < XP_TOP_PLAYERS_MAX_COUNT; t++) {
                int new_top_player = false;
                if (djb_top_players[t] > 0) {
                  int top_player_num = djb_top_players[t] - 1;
                  int top_djb_player_points = djb_player_points[top_player_num];
                  int top_player_level = djb_player_levels[top_player_num];

                  bool cur_player_has_more_pts = (cur_player_points > top_djb_player_points);
                  bool cur_player_is_same_lvl = (cur_player_level == top_player_level);
                  bool cur_player_is_higher_lvl = (cur_player_level > top_player_level);
                  if (
                    (cur_player_has_more_pts && cur_player_is_same_lvl)
                    ||
                    (cur_player_is_higher_lvl)
                  ) {
                    new_top_player = true;
                  }
                }
                else {
                  new_top_player = true;
                }

                if (new_top_player == true) {
                  int t_num = t+1;
                  int next_t = t+1;
                  int next_t_num = next_t+1;
                  for (int tt_num = XP_TOP_PLAYERS_MAX_COUNT; tt_num >= t_num; tt_num--) {
                    int tt = tt_num-1;
                    int prev_tt = tt-1;
                    djb_top_players[tt] = djb_top_players[prev_tt];
                  }

                  djb_top_players[t] = cur_player_num_one_based;

                  t = XP_TOP_PLAYERS_MAX_COUNT;
                }
              }
            }
          }

          // Loop through # of top slots.
          for (t = 0; t < XP_TOP_PLAYERS_MAX_COUNT; t++) {
            if (djb_top_players[t] > 0) {
              cur_player_num_one_based = djb_top_players[t];
              int cur_player_num = cur_player_num_one_based - 1;

              // New line
              hud_message = StrParam(s:hud_message, s:"\n\cj");

              // Player points #
              hud_message = StrParam(s:hud_message, i:djb_player_points[cur_player_num]);

              // Divider (spacer)
              hud_message = StrParam(s:hud_message, s:" \cdL");

              // Player level #
              hud_message = StrParam(s:hud_message, i:djb_player_levels[cur_player_num]);

              // Divider (spacer)
              hud_message = StrParam(s:hud_message, s:" \c-");

              // Player name
              hud_message = StrParam(s:hud_message, n:cur_player_num_one_based);
            }
          }

          // Total points label.
          hud_message = StrParam(s:hud_message, s:"\n\n\cf");

          // Total points #.
          hud_message = StrParam(s:hud_message, i:djb_cur_total_points);

          if (djb_cur_total_levels > 0) {
            DisplayTopPlayersHUD(hud_message);
            Delay(BySeconds(XP_TOP_PLAYERS_HUD_DELAY)); // Use 30.
          }
          else { Delay(10); }

          Restart;
        }
      }

      script SCR_OPEN_3 OPEN {
        ad_map_mon_max_limit = GetCVAR("ad_mon_max_limit");
        ad_remove_corpse_delay = GetCVAR("ad_remove_corpse_delay");

        // If Game is eligible for monster limiting.
        if (IsGameEligibleForMonLimiting()) {
          int m_count, mci, mti, cycle_count, mc_cycle_count;
          int m_base_tid;
          bool mta;

          // If too many monsters in map.
          while (IsMonsterCountTooHigh()) {

            // Prepare monster tid array.
            /*
            for (mci = 0; mci < MON_CLASS_MAX_COUNT; mci++) {
              // Set base tid.
              ad_mon_tid[mci] = ad_mon_base_tid[mci];
            }
            */

            // Loop through each monster class.
            for (mci = 0; mci < MON_CLASS_MAX_COUNT; mci++) {
              if (ad_mon_count[mci] > 0) {
                mta = false;
                // Loop through tid range, and remove every odd/even monster.
                m_base_tid = ad_mon_base_tid[mci] + 1;
                //printbold(s:"m_base_tid = ",i:m_base_tid,s:" - ad_mon_tid[",i:mci,s:"] = ",i:ad_mon_tid[mci]);
                for (mti = m_base_tid; mti <= ad_mon_tid[mci]; mti++) {
                  if (IsMonsterCountTooHigh()) {
                    // mti = monster tid increment.
                    if (mta) {
                      // If monster is alive, and monster class count reduction limit not reached.

                      // Method A: Prioritize lower classes
                      //if (IsActorAlive(mti) && ad_mon_count[mci] > MAP_MON_CLASS_MAX_LIMIT) {

                      // Method B: No priorize.  Overall class limitation.
                      if (
                        ad_mon_count[mci] > ad_map_mon_max_limit
                        ||
                        (ad_mon_count[mci] > MAP_MON_CLASS_MAX_LIMIT && cycle_count >= 1)
                      ) {
                        if (IsActorAlive(mti)) {
                          Thing_Remove(mti);
                          ad_mon_count_total--;
                          ad_mon_count[mci]--;
                          mta = false;
                        }
                      }
                      //printbold(s:"removed ",i:mti);
                    }
                    else {
                      mta = true;
                      //printbold(s:"skipped ",i:mti);
                    }
                  }
                }
              }

              mc_cycle_count++;

              Delay(1);
            }

            cycle_count++;
          }
        }
      }

  /* **************************************************************************
  // ENTER
  // *************************************************************************/
    /*
     * Similar to OPEN scripts, these run once per player, and are always
     * guaranteed to run. They often contain Thing_ChangeTID to give the player
     * a TID, and other player-specific scripts such as custom HUDs. The player
     * is TID 0 in an ENTER script, and this is the first point you can access
     * the player's TID with certainty, as there is no way to set it whilst
     * building the map.
     *
     * Scripts can normally be terminated by the action special ACS_Terminate.
     * However, this is not true for ENTER scripts. If you chose to have an
     * ENTER script running in an infinite loop (with restart at the end, or a
     * while loop, or other methods) then there is no way to stop that script
     * with ACS_Terminate. You can, however, terminate the ENTER script with
     * the terminate keyword called from within the script.
     *
     * Note: In Zandronum, spectators never trigger an ENTER script.
     */

    // 1. Gives player a TID.
    // 2. Sets player as alive.
    script SCR_ENTER ENTER {
      // If Gamemode is NOT deathmatch.
      if (!IsGameTypeDeathmatch()) {
        int p = PlayerNumber();

        Delay(10);

        // Acquire or assign new TID to player, to our djb_player_tids array.
        InitializePlayerTID();

        // Set player as alive.
        PlayerIsAlive(p);

        // Spawn player.
        SpawnPlayerInArena(p);

        // Force our own TID for player.
        int reinit_count = 0;
        while (reinit_count < 20) {
          ReInitializePlayerTID();
          Delay(BySeconds(1));
          reinit_count++;

        }
      }
    }

    // Spawn player in arena.
    script SCR_ENTER_2 (int p) {
      int p_tid = GetPlayerTIDByNum(p);

      // Invasion?  Lives?  Teleport to spectator room.
      // ...
      // Else
      // ...

      // Spawn tear.
      PlayerSpawnTear();

      // Tear echo.
      Delay(BySeconds(1));
      if (IsPlayerPreviouslyDied(p)) {
        ThingSound(p_tid, "Player/Reentry", 127);
      }
      else {
        ThingSound(p_tid, "Player/Entry", 127);
      }

      // Backpack
      if (IS_BACKPACK_SPAWNERS_ENABLED) {
        // Delay.
        Delay(p); // Reduce packet surge.

        // Setup a Backpack spawner at Player #'s spawnpoint.
        StartBackpack();
      }
    }

    // Backpack death (respawn) script.
    script SCR_BACKPACK_RESPAWN (int p) {
      Delay(BySeconds(BP_RESPAWN_DELAY));

      SpawnBackpack(p);
    }

  /* **************************************************************************
  // RETURN
  // *************************************************************************/
    /*
     * Like ENTER scripts, these run once per player and are guaranteed to run.
     * Unlike ENTER scripts, they will run every time the player (re-)enters the
     * level, only if the level has been previously visited (ENTER scripts will
     * run the first time, RETURN will run subsequent times). This type of script
     * is usually used if you need to perform some kind of repeated check every
     * time the player enters the level.
     */
    script SCR_RETURN RETURN { printbold(s:"SCR RETURN"); }

  /* **************************************************************************
  // DEATH
  // *************************************************************************/
    /*
     * Executes when a player dies. The player who died is the script activator, so
     * their number can be determined in a multiplayer game using the PlayerNumber
     * function.
     */

    // 1. Sets player as dead.
    script SCR_DEATH DEATH {
      // If Gamemode is NOT deathmatch.
      if (!IsGameTypeDeathmatch()) {
        int p = PlayerNumber();

        // Player died.
        PlayerDied(p);

        // Hell laugh.
        ACS_Execute(SCR_DEATH_2, 0, 0, 0, 0);

        /*
         * Remove Player's TID during death.
         *
         * No, because we want players to receive their rewards while dead.
         * Switch the TID in RESPAWN event, instead.
         */
        //Thing_ChangeTID(0, 0);
      }
    }

    // Hell laugh.
    script SCR_DEATH_2 (void) {
      // If more than 1 player died within 1 second.
      Delay(BySeconds(1));
      if (djb_recent_p_death_count > EXCITED_LAUGH_DEATH_COUNT) {
        // Excited laugh.
        AmbientSound("Hell/DeathExcitedLaugh", 100);
      }
      // Only 1 player died.
      else {
        // Standard laugh.
        AmbientSound("Hell/DeathStandardLaugh", 100);
      }

      // Wait for laugh sound to finish playing.
      Delay(BySeconds(5));

      // Reset recent player death counter.
      djb_recent_p_death_count = 0;
    }

  /* **************************************************************************
  // RESPAWN
  // *************************************************************************/
    // 1. Resets player's TID.
    // 2. Sets player as alive.
    // 3. Sets player as in-game.
    // 4. Sets player as in-base.
    // 5. Reset player's lives.
    script SCR_RESPAWN RESPAWN {
      Delay(1);
      // If Gamemode is NOT deathmatch.
      if (!IsGameTypeDeathmatch()) {
        // Re-initialize Player's TID.
        ReInitializePlayerTID();

        // Spawn tear.
        PlayerSpawnTear();
      }
    }

  /* **************************************************************************
  // LIGHTNING
  // *************************************************************************/
    /*
     * After the lightning effect occurs, this type of script is run. It is very
     * uncommon to see this type of script and it has one very obvious use.
     */
    // script SCR_LIGHTNING LIGHTNING { printbold(s:"SCR LIGHTNING"); }

  /* **************************************************************************
  // UNLOADING
  // *************************************************************************/
    /*
     * This is almost the inverse to OPEN. It calls when the level is exiting.
     *
     * See: http://zdoom.org/wiki/Script_types#UNLOADING
     */
    //script SCR_UNLOADING UNLOADING { printbold(s:"SCR UNLOADING"); }

  /* **************************************************************************
  // DISCONNECT
  // *************************************************************************/
    /*
     * A DISCONNECT script must be declared with a single argument. ZDoom will pass
     * the number of the disconnecting player into this variable. Because the
     * player has already left the game by the time this script is called, no
     * actions can be taken on that player. The script is executed by the world
     * itself.
     *
     * Note: In Zandronum, DISCONNECT scripts are also executed when a player turns
     * into a spectator.  However, DISCONNECT script is not re-executed when a
     * spectator really disconnects from the server.
     */
    script SCR_DISCONNECT (int p) DISCONNECT {
      // If Gamemode is NOT deathmatch.
      if (!IsGameTypeDeathmatch()) {
        ClearPlayer(p);

        // Update player alive count, using a real check, as a pre-caution.
        djb_player_alive_count = GetRealPlayerAliveCount();

        // If all players left.
        if (GetPlayerCount() < 1) {
          ACS_Execute(SCR_NO_PLAYERS, 0, 0, 0, 0);
        }
      }
    }

    // All players left.  Restarting... (count-down timer)
    script SCR_NO_PLAYERS (void) {
      SetActivator(0);

      int countdown_cursec = RESTART_COUNTDOWN_MAXTIME;
      while (GetPlayerCount() < 1 && countdown_cursec >= 1) {

        SetHudSize(640,480,1); // Sets bar to behave as if it were on a 640x480 screen
        SetFont("BigFont");
        HudMessageBold(s:"No Players Found"; 0, 100000, CR_RED, 320.0, 25.0, 0);
        SetFont("SmallFont");
        HudMessageBold(s:"Restarting in ",i:countdown_cursec,s:"..."; 0, 100001, CR_ORANGE, 320.0, 40.0, 0);

        Delay(TICS_PER_SEC*1);
        countdown_cursec -= 1;
      }
      if (GetPlayerCount() < 1) {
        PrintBold(s:"All players have left.  Restarting...");
        ConsoleCommand("MAP MAP01"); // Full reload.
      }
      else {
        HudMessageBold(s:""; 0, 100000, CR_RED, 320.0, 25.0, 0.1);
        HudMessageBold(s:""; 0, 100001, CR_ORANGE, 320.0, 40.0, 0.1);
      }
    }

  /* ****************************************************************
  // CORE SCRIPTS
  // ***************************************************************/
    script SCR_UNIQUE_TID (void) {
      SetResultValue(UniqueTID());
    }

  /* **************************************************************************
  // SCRIPT: MONSTER SPAWN
  // *************************************************************************/
    // Change monster TID.
    script SCR_MON_SPAWN (int m_type) {
      Delay(Random(1,5));
      ChangeMonsterTIDByClassType(m_type);
    }

  /* **************************************************************************
  // SCRIPT: MONSTER DEATH
  // *************************************************************************/
    // 1. Give points to player(s).
    script SCR_MON_DEATH (int m_type) {
      Delay(Random(1,5));

      // Monster's original health.
      //int mon_orig_health = GetActorProperty(0, APROP_SpawnHealth);
      int mon_orig_health = ad_mon_hp[(m_type-1)];

      // Monster's coordinates.
      int mx = GetActorX(0);
      int my = GetActorY(0);
      int mz = GetActorZ(0);

      // Monster's TID.
      int m_tid = ActivatorTID();
      if (!m_tid) { m_tid = GetNewMonsterTID(); Thing_ChangeTID(0, m_tid); }

      // Change activator to Player.
      SetActivatorToTarget(0);

      // Player's number and TID.
      int p = PlayerNumber();

      // If the killer player is a Player.
      if (p >= 0) {
        int p_tid = GetPlayerTIDByNum(p);

        // If player(s) are still in-game.
        if (IsAnyPlayerJoined()) {
          // Points amount.
          int points = XP_GIVEAWAY_AMT_DEFAULT;

          // Determine amount of points to give, based on monster difficulty.
          // 1. Monster health.
          if (mon_orig_health) {
            // Points based on monster's health.
            points = GetPointsByMonHealth(mon_orig_health); // 10% of HP.

            // Double points for elite monster.
            if (mon_orig_health >= ELITE_MIN_HEALTH) { points *= 2; }
          }

          // 2. Monster speed.
          // 3. Monster power.

          // Give player's the points.
          GivePointsToPlayersAndTarget(points, p);

        }
      }

      // Corpse removal.
      if (IsMonsterCorpseRemovalEnabled()) {
        int new_mon_tid = GetUniqueTID();
        Thing_ChangeTID(m_tid, new_mon_tid);
        int delay_counter = 0;

        // Corpse delay.
        while (
          !IsActorAlive(new_mon_tid)
          &&
          delay_counter <= GetMonsterCorpseRemovalDelay()
        ) {
          Delay(TICS_PER_SEC*1);
          delay_counter++;
        }

        // Remove corpse, if truly dead.
        if (!IsActorAlive(new_mon_tid)) { Thing_Remove(new_mon_tid); }
      }
      else {
        // Remove TID from monster.
        Thing_ChangeTID(m_tid, 0);
      }
    }

  /* **************************************************************************
  // SCRIPT: LEVEL PLAYER
  // *************************************************************************/
    // 1. Give level to player.
    script SCR_LEVEL_PLAYER (int p) {
      int p_tid = GetPlayerTIDByNum(p);
      int p_class_num = GetPlayerClassByNum(p);
      int p_class_num_one_based = p_class_num + 1;
      SetActivator(p_tid);

      int p_num_one_based = p+1;
      int p_xp = CheckActorInventory(p_tid, XP_CLASSNAME); // Not used.
      int p_level = CheckActorInventory(p_tid, XP_LVL_CLASSNAME);
      int p_next_level = p_level + 1;
      int p_next_level_idx = p_next_level-1;

      // Max XP
      int p_max_xp = GetMaxXPByLevelNum(p_next_level);

      // Remaining (difference) XP
      int p_xp_diff = p_xp - p_max_xp; // Not used.

      // Take from Player's points.
      TakeActorInventory(p_tid, XP_CLASSNAME, p_max_xp);

      // Set Player to Level +1
      GiveActorInventory(p_tid, XP_LVL_CLASSNAME, 1);

      // If max level, then reset player's points.
      if (p_next_level >= GetPlayerMaxLevelByNum(p)) {
        TakeActorInventory(p_tid, XP_CLASSNAME, 99999);
      }

      // Give Player full health.
      int p_spawn_health = GetActorProperty(p_tid, "SpawnHealth");
      GiveActorInventory(p_tid, "Health", p_spawn_health);

      // Give Player rewards.
      str reward_weapon_class = GetRewardWeaponClass(p_class_num_one_based, p_next_level);
      int reward_weapon_amt = GetRewardWeaponAmt(p_class_num_one_based, p_next_level);
      // If reward exists.
      if (reward_weapon_amt > 0) {
        str reward_ammo_class = GetRewardAmmoClass(p_class_num_one_based, p_next_level);
        int reward_ammo_amt = GetRewardAmmoAmt(p_class_num_one_based, p_next_level);

        int use_inventory = false;

        /*
         * If player doesn't already have the rewarded weapon, then flag player
         * for auto-equip upon receiving the rewarded weapon.
         */
        if (!CheckActorInventory(p_tid, reward_weapon_class)) {
          use_inventory = true;
        }

        // Give player rewards.
        GiveActorInventory(p_tid, reward_weapon_class, reward_weapon_amt);
        GiveActorInventory(p_tid, reward_ammo_class, reward_ammo_amt);
         IncreaseMaximumHealth(p);
        // If player flagged for auto-equip.
        Delay(1);
        if (use_inventory) {
          // Auto equip the reward.
          UseActorInventory(p_tid, reward_weapon_class);
        }
      }

      // Adjust player's damage factor, to make monsters hit harder.
      /*
      int reward_dmg_factor = XP_LVL_DMG_FACTOR * p_next_level;
      if (p_next_level > 1) { // Skip a factor because level 1.
        reward_dmg_factor = reward_dmg_factor - XP_LVL_DMG_FACTOR;
      }
      if (reward_dmg_factor <= XP_MAX_LVL_DMG_FACTOR) {
        int dmg_factor = FixedDiv(reward_dmg_factor, 100) - 1.0;
        SetActorProperty(p_tid, APROP_DamageFactor, dmg_factor);
      }
      */
      DecreasePlayerDamageFactor(p, XP_LVL_DMG_FACTOR);

      // Play level-up sound for all other players, at half volume.
      OtherPlayersLocalAmbientSound(p, "player/levelup", 64);

      // Player level-up sound for player.
      PlayerLocalAmbientSound(p, "player/levelup", 128);

      // Display level-up HUD.
      for (int c = 0; c < MAX_CLIENT_COUNT; c++) {
        if (PlayerInGame(c)) {
          DisplayLevelUpBarHUDPerPlayer(c, p, p_next_level);

          // if (c == p) {
          //   Log(n:p_num_one_based,s:" has reached level ",i:p_next_level,s:".");
          // }
        }
      }

      // Player level-up effects.
      //SCR_LEVEL_PLAYER_FX
      int fxc, fxci, fxbr;
      str fxc_class;

      // Charge.
      for (fxc = 1; fxc <= 5; fxc++) {
        fxc_class = StrParam(s:"djbLevelUpCharge", i:fxc);
        for (fxci = 0; fxci < 7; fxci++) {
          SpawnSpotForced(fxc_class, p_tid, 0, 0);
          Delay(2);
        }
      }

      // Boom
      SetActivator(0);
      SpawnSpotForced("djbLevelUpBoom", p_tid, 0, 0);
      SpawnSpotForced("djbLevelUpBoomExplosion", p_tid, 0, 0);

      // Boom resin.
      for (fxbr = 0; fxbr < 3; fxbr++) {
        SpawnSpotForced("djbLevelUpBoomResin", p_tid, 0, 0);
        Delay(10);
      }

      // De-charge.
      for (fxc = 5; fxc >= 1; fxc--) {
        fxc_class = StrParam(s:"djbLevelUpCharge", i:fxc);
        for (fxci = 0; fxci < 3; fxci++) {
          SpawnSpotForced(fxc_class, p_tid, 0, 0);
          Delay(2);
        }
      }
    }

  /* **************************************************************************
  // SCRIPT: SHOT SPREAD
  // *************************************************************************/
    script SCR_SHOT_SPREAD (void) {
      int hSpeed = fsqrt(
        FixedMul(GetActorVelX(0), GetActorVelX(0))
        + FixedMul(GetActorVelY(0), GetActorVelY(0))
      );
      int pitch = VectorAngle(hSpeed, GetActorVelZ(0))*360 + Random(-7.1, 7.1);

      SetResultValue(pitch);
    }



/*###########################################################################*/
/* FUNCTIONS ################################################################*/
/*###########################################################################*/

  /* **************************************************************************
  // CORE FUNCTIONS
  // *************************************************************************/
    function int abs(int input) {
      if (input < 0) { return -input; }
      return input;
    }

    function int msqr(int n) { return n*n; }

    // Copied from ZDWiki.
    // Fixed-point version, only works up to 30,000 or so:
    function int fsqrt(int number) {
      int samples = 15; // Samples for accuracy

      if (number == 1.0) { return 1.0; }
      if (number <= 0) { return 0; }
      int val = samples<<17 + samples<<19; //x*10 = x<<1 + x<<3
      for (int i=0; i<samples; i++) {
        val = (val + FixedDiv(number, val)) >> 1;
      }

      return val;
    }

    function int fdistance(int tid1, int tid2) {
      int len;
      int y = getactory(tid1) - getactory(tid2);
      int x = getactorx(tid1) - getactorx(tid2);
      int z = getactorz(tid1) - getactorz(tid2);

      int ang = vectorangle(x,y);
      if (((ang+0.125)%0.5) > 0.25) len = fixeddiv(y, sin(ang));
      else len = fixeddiv(x, cos(ang));

      ang = vectorangle(len, z);
      if (((ang+0.125)%0.5) > 0.25) len = fixeddiv(z, sin(ang));
      else len = fixeddiv(len, cos(ang));

      return len;
    }

    function int int2fix(int value)     { return value * 1.0; }
	
function void SetXPRequired (void)
{
	int XPRequired = (CheckInventory("DJBLevel")*100)+(CheckInventory("DJBLevel")*GameSkill()*25);
	SetAmmoCapacity ("djbPoint", XPRequired);
}

  /* **************************************************************************
  // HELPER FUNCTIONS
  // *************************************************************************/
    function bool IsGameTypeInvasion(void) {
      if (GetCVAR("invasion")) { return true; }
      return false;
    }

    function bool IsGameTypeSurvival(void) {
      if (GetCVAR("sv_maxlives") > 0) { return true; }
      return false;
    }

    function bool IsGameTypeSinglePlayer(void) {
      if (GameType() == GAME_TITLE_MAP || GameType() == GAME_SINGLE_PLAYER) {
        return true;
      }
      return false;
    }

    function bool IsGameTypeTitle(void) {
      if (GameType() == GAME_TITLE_MAP) {
        return true;
      }
      return false;
    }

    function bool IsGameTypeDeathmatch(void) {
      if (GameType() == GAME_NET_DEATHMATCH) { return true; }
      return false;
    }

    function bool IsGameTypeMultiplayerCoop(void) {
      return (GameType() == GAME_NET_COOPERATIVE);
    }

    function int BySeconds(int sec) {
      return TICS_PER_SEC*sec;
    }

    function int GetTimer(void) {
      return (Timer()/TICS_PER_SEC);
    }

    function int GetLevelNum(void) {
      return GetLevelInfo(LEVELINFO_LEVELNUM);
    }

    function int GetUniqueTID(void) { return UniqueTID(); }

    function bool IsActorAlive(int actor_tid) {
      if (GetActorProperty(actor_tid, APROP_HEALTH) <= 0) { return false; }
      return true;
    }

    function int GetActorDistance(int source, int target) {
      if (
        ThingCount(T_NONE, target) == 0
        || ThingCount(T_NONE, source) == 0
      ) { return false; }

      int distance = msqr(GetActorX(source) - GetActorX(target)>>16);
      distance += msqr(GetActorY(source) - GetActorY(target)>>16);
      distance += msqr(GetActorZ(source) - GetActorZ(target)>>16);

      return distance;
    }

    function bool CheckActorDistance(int source, int target, int range) {
      if (GetActorDistance(source, target) <= msqr(range)) { return true; }

      return false;
    }

    function void Thing_HateAllPlayersAndTarget(int hater_tid, int hatee_tid, int hate_type) {
      // All players, except target.
      for (int p = 0; p < MAX_CLIENT_COUNT; p++) {
        int p_tid = GetPlayerTIDByNum(p);
        if (IsPlayerInGame(p) && p_tid != hatee_tid) {
          Thing_Hate(hater_tid, GetPlayerTIDByNum(p), hate_type);
        }
      }
      // Hate target.
      Thing_Hate(hater_tid, hatee_tid, hate_type);
    }

  /* **************************************************************************
  // PLAYER FUNCTIONS
  // *************************************************************************/
    function void ClearPlayerTID(int p) { djb_player_tids[p] = 0; }
    function void ClearPlayerDeaths(int p) { djb_player_deaths[p] = 0; }
    function void ClearPlayerLevels(int p) { djb_player_levels[p] = 0; }
    function void ClearPlayerPoints(int p) { djb_player_points[p] = 0; }
    function void ClearPlayerAlive(int p) { PlayerIsNotAlive(p); }
    function void ClearPlayer(int p) {
      ClearPlayerTID(p); // djb_player_tids[p] = 0;
      ClearPlayerDeaths(p); // djb_player_deaths[p] = 0;
      ClearPlayerLevels(p); // djb_player_levels[p] = 0;
      ClearPlayerPoints(p); // djb_player_points[p] = 0;
      ClearPlayerAlive(p); // PlayerIsNotAlive(p);
    }

    function void KillPlayer(int p) {
      int p_tid = GetPlayerTIDByNum(p);

      SetActorProperty(p_tid, APROP_Invulnerable, OFF);

      // Remove armor, so that our slight damage of "10" will work(kill the player)
      TakeActorInventory(p_tid, "Armor", 999);

      // Set Health to 1, so that our slight damage of "DJV" is enough to kill player.
      SetActorProperty(p_tid, APROP_Health, 1);

      // Damage player by 10, to kill player.
      Thing_Damage2(p_tid, 10, 0);

      // Fix bug(rarely occurs) where player corpse still has "1" health.  Force corpse ot have 0 health, so boss stops moving/following the corpse.
      SetActorProperty(p_tid, APROP_Health, 0);
      // TODO: For single-player games, this may not be working correct.  Try destroying the player instead.
    }

    function void PlayerDied(int p) {
      // Set player as not alive.
      PlayerIsNotAlive(p);

      // Add to death count.
      djb_player_deaths[p]++;
      djb_recent_p_death_count++;
    }

    function int GetPlayerCount(void) { return PlayerCount(); }
    function bool IsAnyPlayerJoined(void) { return (GetPlayerCount() > 0); }

    function int GetPlayerDeathCount(int p) {
      return djb_player_deaths[p];
    }

    function int IsPlayerPreviouslyDied(int p) {
      return (GetPlayerDeathCount(p) > 0);
    }

    function int GetRealPlayerAliveCount(void) {
      int player_count = PlayerCount();
      int alive_player_count = 0;
      for (int p = 0; p < MAX_CLIENT_COUNT; p++) {
        if (IsRealPlayerAliveByNum(p)) { alive_player_count++; }
      }
      return alive_player_count;
    }
    function int GetPlayerAliveCount(void) { return djb_player_alive_count; }

    function bool IsRealPlayerAliveByNum(int p) {
      return (
        IsPlayerInGame(p)
        &&
        IsActorAlive(GetPlayerTIDByNum(p))
      );
    }
    function bool IsPlayerAliveByNum(int p) { return (djb_player_alive[p]); }

    function bool IsAnyPlayerAlive(void) { return (GetPlayerAliveCount() > 0); }

    function void IncreasePlayerAliveCount(void) { djb_player_alive_count++; }
    function void DecreasePlayerAliveCount(void) { djb_player_alive_count--; }

    function void PlayerIsAlive(int p) {
      if (!IsPlayerAliveByNum(p)) { IncreasePlayerAliveCount(); }
      djb_player_alive[p] = true;
    }

    function void PlayerIsNotAlive(int p) {
      if (IsPlayerAliveByNum(p)) { DecreasePlayerAliveCount(); }
      djb_player_alive[p] = false;
    }

    function int GetRandomAlivePlayerTID(void) {
      int min_random_player = 1;
      int max_random_player = GetPlayerAliveCount(); // PlayersOutsideCount();
      int random_player = 1;
      if (max_random_player > min_random_player) {
        random_player = random(min_random_player, max_random_player);
      }
      int player_found_count = 0;
      int random_player_tid = GetPlayerTIDByNum(0);
      for (int p = 0; p < 63; p++) {
        //if (!IsPlayerInBase(p) && IsPlayerAliveByNum(p)) {
        if (IsPlayerAliveByNum(p)) {
          player_found_count++;
          if (player_found_count == random_player) {
            random_player_tid += p;
            break;
          }
        }
      }
      return random_player_tid;
    }

    function bool IsPlayerInGame(int p) { return PlayerInGame(p); }

    function bool IsOnlyOnePlayer(void) { return (PlayerCount() == 1); }

    function bool IsPlayerEligibleForPoints(int p) {
      return IsPlayerInGame(p);
    }

    function void PlayerLocalAmbientSound(int player, str sound, int volume) {
      SetActivator(GetPlayerTIDByNum(player));
      LocalAmbientSound(sound, volume);
    }

    function void OtherPlayersLocalAmbientSound(int player, str sound, int volume) {
      for (int p = 0; p < MAX_CLIENT_COUNT; p++) {
        if (PlayerInGame(p)) {
          if (p != player) { PlayerLocalAmbientSound(p, sound, volume); }
        }
      }
    }

    function void DecreasePlayerDamageFactor(int p, int percent) {
      int p_tid = GetPlayerTIDByNum(p);
      int p_dmgfactor = GetActorProperty(p_tid, APROP_DamageFactor);
      p_dmgfactor -= FixedDiv(percent, 100);
      int p_maxdmgfactor = FixedDiv(XP_MAX_LVL_DMG_FACTOR, 100);
      if (p_dmgfactor < p_maxdmgfactor) { p_dmgfactor = p_maxdmgfactor; }
      SetActorProperty(p_tid, APROP_DamageFactor, p_dmgfactor);
    }
	



    /*
    function void ResetPlayerTimer(int p) {
      djb_player_tear_timer[p] = 0;
    }

    function int GetNewPlayerTimer(void) {
      int multiplier = random(TEAR_TIMER_MULTIPLIER_MIN,TEAR_TIMER_MULTIPLIER_MAX);
      int new_timer = multiplier * TEAR_TIMER_MIN;
      new_timer += GetTimer();
      return new_timer;
    }

    function bool IsPlayerTimerValid(int timer) {
      if (timer > 0) { return true; }
      return false;
    }

    function int GetPlayerTimerByNumber(int p) {
      int p_timer = djb_player_tear_timer[p];
      if (!IsPlayerTimerValid(p_timer)) {
        p_timer = GetNewPlayerTimer();
        djb_player_tear_timer[p] = p_timer;
      }
      return p_timer;
    }
    */

    function void SpawnPlayerInArena(int p) {
      ACS_ExecuteAlways(SCR_ENTER_2, 0, p, 0, 0);
    }

    function void InitializePlayerTID(void) {
      int current_tid = ActivatorTID();

      int p_num = PlayerNumber();
      int player_tid = djb_player_tids[p_num];

      bool is_tid_old = false;

      // If player has no TID.
      if (current_tid == 0) {
        // Generate/change TID.
        current_tid = UniqueTID();
        Thing_ChangeTID(0, current_tid);
        is_tid_old = true;
      }
      // Player already has TID.
      else {
        // If it's different, then consider it as old.
        if (player_tid != current_tid) { is_tid_old = true; }
      }

      // If TID is old (has changed), then save it.
      if (is_tid_old) {
        djb_player_tids[p_num] = current_tid;
      }
    }

    function void ReInitializePlayerTID(void) {
      int p = PlayerNumber();
      int p_tid = ActivatorTID();
      Thing_ChangeTID(p_tid, 0);
      djb_player_tids[p] = 0;
      InitializePlayerTID();
    }

    function int GetPlayerTIDByNum(int p) { return djb_player_tids[p]; }
    function int GetPlayerClassByNum(int p) {
      int p_class = PlayerClass(p);
      if (p_class < 0) { p_class = 0; }
      return p_class;
    }
    function void PlayerSpawnTear(void) {
      int p_tid = ActivatorTID();

      // Spawn tear.
      SpawnSpotForced("PlayerSpawnTear", p_tid, 0, 0);
    }

    function void LevelPlayer(int p) {
      ACS_ExecuteAlways(SCR_LEVEL_PLAYER, 0, p, 0, 0);
    }

  /* **************************************************************************
  // MONSTER FUNCTIONS
  // *************************************************************************/
    function int GetMonsterCorpseRemovalDelay(void) {
      return ad_remove_corpse_delay;
    }

    function bool IsMonsterCorpseRemovalEnabled(void) {
      return (GetMonsterCorpseRemovalDelay() > 0);
    }

    function void ChangeMonsterTIDByClassType(int type) {
      int type_idx = type - 1;
      ad_mon_tid[type_idx]++;
      Thing_ChangeTID(0, ad_mon_tid[type_idx]);
    }

    function int GetMonsterCount(void) {
      if (!ad_mon_count_total) {
        int m_count, mc_count, mci, mti;

        // mci = monster class index #.
        // Loop through all monster classes, and collect counts of monsters.
        for (mci = 0; mci < MON_CLASS_MAX_COUNT; mci++) {
          str m_class = ad_mon_class[mci];
          mc_count = ThingCountName(m_class, 0);
          ad_mon_count[mci] = mc_count; // Save monster class count.
          m_count += mc_count;
        }

        ad_mon_count_total = m_count; // Save monster total count.
      }

      return ad_mon_count_total;
    }

    function int GetMonsterMaxLimit(void) { return ad_map_mon_max_limit; }

    function bool IsMonsterCountTooHigh(void) {
      return (GetMonsterCount() > GetMonsterMaxLimit());
    }

    function bool IsGameEligibleForMonLimiting(void) {
      if (
        !IsGameTypeDeathmatch()
        && !IsGameTypeTitle()
        && GetMonsterMaxLimit() > 0
      ) { return true; }
      return false;
    }

    /*
    function int GetMonOrigHealthBySkill(int g_skill) {
      if (g_skill == 2) {
        mon_orig_health = mon_orig_health / 3;
      }
      else if (g_skill == 3) {
        mon_orig_health = mon_orig_health / 6;
      }
      else if (g_skill == 4) {
        mon_orig_health = mon_orig_health / 9;
      }
    }
    */

  /* **************************************************************************
  // BACKPACK FUNCTIONS
  // *************************************************************************/
    function int GetMaxNumOfBackpackSpawners(void) { return MAX_BP_SPAWNERS; }

    function bool IsBackpacksSpawned(void) { return djb_bps_spawned; }
    function void BackpacksIsSpawned(void) { djb_bps_spawned = true; }

    function void StartBackpack(void) {
      int p = PlayerNumber();
      if (!IsBackpacksSpawned()) {
        /*
         * Immediately set that backpacks are spawned, because we only want
         * this to occur once (for first player joining), and we know the
         * spawns will be successful because we're spawning with force.
         */
        BackpacksIsSpawned();

        // Get Player TID and coordinates.
        int p_tid = GetPlayerTIDByNum(p);
        int p_x = GetActorX(p_tid);
        int p_y = GetActorY(p_tid);
        int p_z = GetActorZ(p_tid);

        // Spawn mapspots for our spawners.
        for (int sc = 0; sc < GetMaxNumOfBackpackSpawners(); sc++) {
          djb_bp_spawner_tids[sc] = UniqueTID();
          djb_bp_tids[sc] = UniqueTID();

          // Generate coordinates for spawning a spawner.
          int bp_ms_x, bp_ms_y;
          if (sc == 1) {
            bp_ms_x = p_x + random(-32.0, -56.0);
            bp_ms_y = p_y + random(-32.0, -56.0);
          }
          else if (sc == 2) {
            bp_ms_x = p_x + random(32.0, 56.0);
            bp_ms_y = p_y + random(32.0, 56.0);
          }
          else if (sc == 3) {
            bp_ms_x = p_x + random(-32.0, -56.0);
            bp_ms_y = p_y + random(32.0, 56.0);
          }
          else {
            bp_ms_x = p_x + random(random(-32.0, -56.0), random(32.0, 56.0));
            bp_ms_y = p_y + random(random(-32.0, -56.0), random(32.0, 56.0));
          }
          int bp_ms_z = p_z + random(32.0, 64.0);

          SpawnForced("MapSpot", bp_ms_x, bp_ms_y, bp_ms_z, djb_bp_spawner_tids[sc], 0);

          // Spawn backpack.
          SpawnBackpack(sc);
        }
      }
    }

    function int GetBackpackSpawnerTIDByNum(int p) {
      return djb_bp_spawner_tids[p];
    }

    function int GetBackpackTIDByNum(int p) {
      return djb_bp_tids[p];
    }

    function void SpawnBackpack(int num) {
      int bp_spawner_tid = GetBackpackSpawnerTIDByNum(num);
      int bp_tid = GetBackpackTIDByNum(num);

      // Spawn backpack at the mapspot.
      SpawnSpotForced(BP_CLASSNAME, bp_spawner_tid, bp_tid, 0);
      SpawnSpot("ItemFog", bp_spawner_tid);

      // Assign special for re-spawning upon pick-up.
      SetThingSpecial(bp_tid, ACS_ExecuteAlways, SCR_BACKPACK_RESPAWN, 0, num);
    }
	

  /* **************************************************************************
  // TEAM FUNCTIONS
  // *************************************************************************/
    /*
     * GetTeamNumByPlayerNum()
     *
     * Which team the player is on.
     * No team is 255 always.
     * Without use of the TEAMINFO lump the teams are:
     * 0 for blue, 1 for red, 2 for green, 3 for gold,
     * 4 for black, 5 for white, 6 for orange and 7 for purple.
     */
    function int GetTeamNumByPlayerNum(int p_num) {
      return GetPlayerInfo(p_num, PLAYERINFO_TEAM);
    }

    function str GetTeamNameByNum(int t_num) {
      return djb_team_name[t_num];
    }

    function str GetTeamNameByPlayerNum(int p_num) {
      return GetTeamNameByNum(GetTeamNumByPlayerNum(p_num));
    }

    function str GetTeamColorByNum(int t_num) {
      return djb_team_color[t_num];
    }

    function str GetTeamColorByPlayerNum(int p_num) {
      return GetTeamColorByNum(GetTeamNumByPlayerNum(p_num));
    }

  /* **************************************************************************
  // CONFIG FUNCTIONS
  // *************************************************************************/
    // Player Class Number, Reward Number.
    function int GetRewardWeaponNum(int rpc_num, int r_num) {
      // Reward -> Player Class -> Reward Number -> Weapon Number
      return djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_REWARD *xb*xc*xd+ rpc_num *xc*xd+ r_num *xd+ IDX_CFG_REWARD_WNUM];
    }

    // Player Class Number, Reward Number.
    function str GetRewardWeaponClass(int rpc_num, int r_num) {
      int w_num = GetRewardWeaponNum(rpc_num, r_num);

      // Weapon -> Player Class -> Weapon Number -> Weapon Name
      return djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_WEAPON *xb*xc*xd+ rpc_num *xc*xd+ w_num *xd+ IDX_CFG_ITEM_WEAPON_CLASS];
    }

    // Player Class Number, Reward Number.
    function int GetRewardWeaponAmt(int rpc_num, int r_num) {
      int w_num = GetRewardWeaponNum(rpc_num, r_num);

      // Weapon -> Player Class -> Weapon Number -> Weapon Amount
      return djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_WEAPON *xb*xc*xd+ rpc_num *xc*xd+ w_num *xd+ IDX_CFG_ITEM_WEAPON_AMT];
    }

    // Player Class Number, Reward Number.
    function int GetRewardAmmoNum(int rpc_num, int r_num) {
      // Reward -> Player Class -> Reward Number -> Weapon Number
      return djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_REWARD *xb*xc*xd+ rpc_num *xc*xd+ r_num *xd+ IDX_CFG_REWARD_ANUM];
    }

    // Player Class Number, Reward Number.
    function str GetRewardAmmoClass(int rpc_num, int r_num) {
      int a_num = GetRewardAmmoNum(rpc_num, r_num);

      // Weapon -> Player Class -> Weapon Number -> Weapon Name
      return djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_AMMO *xb*xc*xd+ rpc_num *xc*xd+ a_num *xd+ IDX_CFG_ITEM_AMMO_CLASS];
    }

    // Player Class Number, Reward Number.
    function int GetRewardAmmoAmt(int rpc_num, int r_num) {
      int a_num = GetRewardAmmoNum(rpc_num, r_num);

      // Weapon -> Player Class -> Ammo Number -> Ammo Amount
      return djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_AMMO *xb*xc*xd+ rpc_num *xc*xd+ a_num *xd+ IDX_CFG_ITEM_AMMO_AMT];
    }

    // XP
    function int GetXpMaxIncrAmt(void) {
      return djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_XP *xb*xc*xd+ 0 *xc*xd+ 0 *xd+ IDX_CFG_XP_MAX_INCR];
    }

    function int GetXpMaxIncrAmtLimit(void) {
      return djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_XP *xb*xc*xd+ 0 *xc*xd+ 0 *xd+ IDX_CFG_XP_MAX_INCR_LIMIT];
    }

  /* **************************************************************************
  // GAME FUNCTIONS
  // *************************************************************************/
    function int GetNewMonsterTID(void) { return UniqueTID(); }

    /*
     * Max Level
     *
     * The maximum level a player can go.
     */
    function int GetPlayerMaxLevelByNum(int p) {
      SetActivator(GetPlayerTIDByNum(p));
      return GetAmmoCapacity(XP_MAX_LVL_CLASSNAME);
    }

    function int GetMaxXPByLevelNum(int level) {
      int level_idx = level-1;
      int max_xp;
      int incr_xp = GetXpMaxIncrAmt();

      if (level > 1) {
        max_xp = incr_xp; // E.g. Add the first one (for Level 1 only.)
        // Now loop through remaining levels, and increase max_xp each time.
        for (int l = 1; l < level; l++) {
          max_xp = max_xp + incr_xp;
        }
      }

      // Enforce max limit.
      if (max_xp > GetXpMaxIncrAmtLimit()) {
        max_xp = GetXpMaxIncrAmtLimit();
      }

      return max_xp;
    }

    function int GetHalfPoints(int points) { return ((points / 2) * 1); }

    function int GetPointsByMonHealth(int mon_hp) {
      int points = mon_hp / XP_GIVEAWAY_PERCENT_BY_MON_HP;
      // Force minimum.
      if (points < XP_GIVEAWAY_AMT_DEFAULT) {
        points = XP_GIVEAWAY_AMT_DEFAULT;
      }
      return points;
    }

    function void GivePointsToPlayersAndTarget(int points, int target_num) {
      // If killer player is not a player, then give no points to no one.
      if (target_num < 0) { return; }

      int p_num; // Dynamic player to read max level from.
      int p_tid; // Dynamic player to give points to. (Used in loop below.)
      int target_tid = GetPlayerTIDByNum(target_num); // The killer player.

      // # of players.
      int num_of_players = GetPlayerCount();

      // Loop through all players.
      for (int p = 0; p < MAX_CLIENT_COUNT; p++) {
        if (IsPlayerEligibleForPoints(p)) {
          int give_points = GetPointsSlice(points, num_of_players);

          // Not the killer, then give half-points.
          if (p != target_num) {
            give_points = GetHalfPoints(give_points);
            p_num = p;
            p_tid = GetPlayerTIDByNum(p);
          }
          else {
            p_num = target_num;
            p_tid = target_tid;
          }

          // If player # is not max level.
          if (CheckActorInventory(p_tid, XP_LVL_CLASSNAME) < GetPlayerMaxLevelByNum(p_num)) {
            // Give points to player #.
            GiveActorInventory(p_tid, XP_CLASSNAME, give_points);

            // Debug
            //PrintBold(s:"Player ",i:p,s:" received ",i:give_points,s:" pts.");
          }
        }
      }
    }

    function int GetPointsSlice(int points, int num_of_players) {
      if (num_of_players == 0) { num_of_players = 1; } // Safety.

      // Setup so that points are evenly distributed amongst all eligible
      // players. (Divide amount of points by # of players.)
      // Plus give the owner full points.
      int dist_points = (points / num_of_players) * 1;

      // Setup so that each player receives a slightly different amount of
      // points.
      int random_points_slice_amt = (dist_points * XP_RAND_PTS_SLICE_PERCENT) / 100;

      // Enforce minimum.
      if (random_points_slice_amt < XP_RAND_PTS_SLICE_AMT_MIN) {
        random_points_slice_amt = XP_RAND_PTS_SLICE_AMT_MIN;
      }

      // Determine start and end points of slice range for random amount.
      int random_points_start = dist_points - random_points_slice_amt;
      int random_points_end = dist_points + random_points_slice_amt;

      // Deliver random amount of points of slice.
      return random(random_points_start, random_points_end);
    }
	
	

  /* **************************************************************************
  // ENGINE FUNCTIONS
  // *************************************************************************/
    function void Game(void) {
      int p_tid;
      int p_inbase_count;

      // Fixed settings.
      /*
      if (GetCVAR("ao_survival_maxlives") > SURVIVAL_MAX_LIVES) {
        SetCVAR("ao_survival_maxlives", SURVIVAL_MAX_LIVES);
      }
      if (GetCVAR("ao_nolivestimeout") != GAMEOVER_COUNTDOWN_TIME) {
        SetCVAR("ao_nolivestimeout", GAMEOVER_COUNTDOWN_TIME);
      }
      */

      // Settings.
      //if (!djb_max_level) { djb_max_level = GetCVAR(CVAR_XP_MAX_LVL); }

      // ------------------------------------------------------------------------

      // Viles HUD.
      str hud_viles = "\n";
      int vile_count = ThingCountName("Archvile2", 0);
      //int exile_count = ThingCountName("T2Exile", 0);
      int exile_count = 0;
      if (vile_count > 0) {
        hud_viles = StrParam(s:hud_viles, s:"Archvile(s):\cc ");
        hud_viles = StrParam(s:hud_viles, i:vile_count);
        hud_viles = StrParam(s:hud_viles, s:"\n");
      }
      // if (exile_count > 0) { ... }
      if (vile_count > 0 || exile_count > 0) {
        SetHudSize(640,480,True);
        SetFont("SmallFont");
        HudMessageBold(s:hud_viles; 0, 4540, CR_RED, 320.4, 20.1, 1.5); // 0.32, 0.34, 0.36, etc.
      }

      // ------------------------------------------------------------------------

      // Check all players.
      for (int p = 0; p < MAX_CLIENT_COUNT; p++) {
        // Player's TID.
        p_tid = GetPlayerTIDByNum(p);

        // Player is eligible for levelling.
        if (IsPlayerEligibleForPoints(p)) {
          // Player's XP and Level.
          int p_xp = CheckActorInventory(p_tid, XP_CLASSNAME);
          int p_level = CheckActorInventory(p_tid, XP_LVL_CLASSNAME);
          int p_next_level = p_level+1;
          if (p_next_level > GetPlayerMaxLevelByNum(p)) { p_next_level = GetPlayerMaxLevelByNum(p); }
          int p_max_xp;
          if (p_level < p_next_level && p_next_level > 1) {
            p_max_xp = GetMaxXPByLevelNum(p_next_level);
            // Player has enough XP to level-up.
            if (p_xp >= p_max_xp) {
              // Level-up player.
              LevelPlayer(p);
            }
          }
        }

        /*
          // Player is eligible for base detection.
          if (IsPlayerEligibleForBaseDetection(p)) {
            if (IsPlayerInBase(p)) {
              p_inbase_count++;
            }
          }
        */

        // Player is eligible for regeneration.
        if (IS_REGEN_ENABLED && IsPlayerAliveByNum(p)) {
          p_tid = GetPlayerTIDByNum(p);
          SetActivator(p_tid);

          /*
           * Item, Rate, Tics
           *
           * Regen rate limit = 5
           * Regen rate increment = 2
           * Repeats every 1 tic.
           *
           * Regen every 4 frames, would mean tics = 4.
           * This wouldn't work with a limit of 5 and increment of 2, because it
           * only takes 3 tics to reach regen amount of 6 which is over the limit.
           * Solution: Change rate limit and increment to work with
           *           4-frame (animation) intervals.
           *           Example: Rate limit = 4, increment = 2
           * But, rate limit and increment are globals that won't change.  So
           * What happens if we're going by 5-frame or 10-frame animation invervals.
           * Pass in #-frame (tics) as an argument to the ACS script, and use math
           * to figure it out.  Although can be accurate, this would still slow
           * down the visuals of the increments.  Player might get confused why
           * suddenly the increments are slower, yet larger.
           *
           * Ultimately, need to stick to 1 standard of #-frame intervals.  The
           * lower, the better.
           *
           * Is executing ACS on every #-frame interval too much?
           */

          for (int t = 0; t < REGEN_MAX_TYPES; t++) {
            str regen_class        = regen_attr[t][0]; // E.g. Stamina
            str regen_type         = regen_attr[t][1]; // E.g. "ammo"
            str regen_rate_class   = regen_attr[t][2]; // E.g. StaminaRegenRate
            str regen_rate_type    = regen_attr[t][3]; // E.g. "ammo"
            str regen_amount_class = regen_attr[t][4]; // E.g. StaminaRegenAmt
            str regen_amount_type  = regen_attr[t][5]; // E.g. "ammo"

            int regen_max;
            switch(regen_type) {
              case REGEN_TYPE_AMMO:
                regen_max = GetAmmoCapacity(regen_class);
                break;
              case REGEN_TYPE_INV:
                //regen_max = GetMaxInventory(regen_class);
                break;
            }

            int regen_rate;
            switch(regen_rate_type) {
              case REGEN_TYPE_AMMO:
                regen_rate = CheckInventory(regen_rate_class);
                break;
              case REGEN_TYPE_INV:
                //regen_rate = GetMaxInventory(regen_rate_class);
                break;
            }

            int regen_amount_limit;
            switch(regen_amount_type) {
              case REGEN_TYPE_AMMO:
                regen_amount_limit = GetAmmoCapacity(regen_amount_class);
                break;
              case REGEN_TYPE_INV:
                //regen_amount_limit = GetMaxInventory(regen_amount_class);
                break;
            }

            // If item is not full.
            // E.g. If "Stamina" < 100.
            if (CheckInventory(regen_class) < regen_max) {
              // If item's regen rate limit hasn't been reached yet.
              // E.g. If "StaminaRegenAmt" < 10.
              if (CheckInventory(regen_amount_class) < regen_amount_limit) {
                // Add/increment rate amount.
                // E.g. Give "StaminaRegenAmt" 1 ("StaminaRegenRate").
                GiveInventory(regen_amount_class, regen_rate);
              }
              // Item's regen rate limit reached.
              else {
                // Clear rate amount.
                TakeInventory(regen_amount_class, regen_amount_limit);

                if (CheckInventory(regen_class) < regen_max) {
                  GiveInventory(regen_class, 1);
                }
              }
            }
          }
        }
      }

      /*
        // Update player_inbase_count.
        SetPlayerInBaseCount(p_inbase_count);

        // If all players in base or no players playing.
        if (PlayerCount() > 0 && PlayerInBaseCount() >= PlayerCount() ) {
          // If gameover timer not started.
          if (!IsGameOverTimer()) {
            // Start 30-second countdown for free joining.
            ACS_Execute(59, 0, 0, 0, 0);
          }
        }
        else {
          // Cancel gameover timer.
          if (IsGameOverTimer()) { gameover_timer = false; }
        }
      */
    }

    function void DisplayNoPlayersMessageHUD(void) {
      SetHudSize(640,480,1);
      SetFont("BigFont");

      HudMessageBold(
        l:HUD_NO_PLAYERS_TEXT;
        HUD_NO_PLAYERS_TYPE,
        HUD_NO_PLAYERS_CHAN,
        HUD_NO_PLAYERS_COLOR,
        HUD_NO_PLAYERS_X,
        HUD_NO_PLAYERS_Y,
        HUD_NO_PLAYERS_DURATION
      );
    }

    function void ClearNoPlayersMessageHUD(void) {
      SetHudSize(640,480,1);
      SetFont("BigFont");

      HudMessageBold(
        s:"";
        HUD_NO_PLAYERS_TYPE,
        HUD_NO_PLAYERS_CHAN,
        HUD_NO_PLAYERS_COLOR,
        HUD_NO_PLAYERS_X,
        HUD_NO_PLAYERS_Y,
        0.1
      );
    }

    function void DisplayTopPlayersHUD(str msg) {
      SetHudSize(640,480,1);
      SetFont("SmallFont");
      HudMessageBold(
        s:msg;
        HUD_PLAYERS_SCORE_TYPE,
        HUD_PLAYERS_SCORE_CHAN,
        HUD_PLAYERS_SCORE_COLOR,
        HUD_PLAYERS_SCORE_X,
        HUD_PLAYERS_SCORE_Y,
        HUD_PLAYERS_SCORE_DURATION,
        HUD_PLAYERS_SCORE_FADETIME
      );
    }

    function void DisplayLevelUpBarHUDPerPlayer(int c, int p, int level) {
      bool theirs = (c != p);

      str txt_you = strparam(l:HUD_LUP_TXT_WORD_YOU, s:""); // "You"
      str txt_reached = strparam(s:" ", l:HUD_LUP_TXT_WORD_REACHED); // " reached"
      str txt_level = strparam(l:HUD_LUP_TXT_WORD_LEVEL, s:" "); // "Level "
          txt_level = strparam(s:txt_level, i:level);
      int txt_level_channel = HUD_LUP_TXT_LEVEL_CHAN + c;
      int txt_level_color = HUD_LUP_TXT_LEVEL_COLOR;
      int txt_level_y = HUD_LUP_TXT_LEVEL_Y;

      str bar_font = HUD_LUP_FRAME_FONT;
      int bar_channel = HUD_LUP_FRAME_CHAN + c;
      int bar_y = HUD_LUP_FRAME_Y;

      str txt_reached_font = HUD_LUP_TXT_REACHED_FONT;
      int txt_reached_channel = HUD_LUP_TXT_REACHED_CHAN + c;
      int txt_reached_y = HUD_LUP_TXT_REACHED_Y;

      if (theirs) {
        int p_num_one_based = p + 1;
        txt_reached = strparam(n:p_num_one_based, s:txt_reached);

        bar_font = HUD_THEIR_LVLUP_FRAME_FONT;
        bar_channel = HUD_THEIR_LVLUP_FRAME_CHAN;
        bar_y = HUD_THEIR_LVLUP_FRAME_Y;

        txt_reached_font = HUD_THEIR_LUP_TXT_REACHED_FONT;
        txt_reached_channel = HUD_THEIR_LUP_TXT_REACHED_CHAN;
        txt_reached_y = HUD_THEIR_LUP_TXT_REACHED_Y;

        txt_level_channel = HUD_THEIR_LUP_TXT_LEVEL_CHAN;
        txt_level_color = HUD_THEIR_LUP_TXT_LEVEL_COLOR;
        txt_level_y = HUD_THEIR_LUP_TXT_LEVEL_Y;
      }
      else {
        txt_reached = strparam(s:txt_you, s:txt_reached);
      }

      SetActivator(GetPlayerTIDByNum(c));

      SetHudSize(640,480,1);

      // Bar
      SetFont(bar_font); // HUD_LUP_FRAME_FONT || HUD_THEIR_LVLUP_FRAME_FONT
      HudMessage(
        s:HUD_LUP_FRAME_TEXT;
        HUD_LUP_FRAME_TYPE,
        bar_channel,
        HUD_LUP_FRAME_COLOR,
        HUD_LUP_FRAME_X,
        bar_y,
        HUD_LUP_FRAME_DURATION,
        HUD_LUP_FRAME_FADETIME,
        HUD_LUP_FRAME_FADEOUTTIME);

      // Draw "You've reached" HUD.
      SetFont(txt_reached_font);
      HudMessage(
        s:txt_reached;
        HUD_LUP_FRAME_TYPE,
        txt_reached_channel,
        HUD_LUP_TXT_REACHED_COLOR,
        HUD_LUP_FRAME_X,
        txt_reached_y,
        HUD_LUP_FRAME_DURATION,
        HUD_LUP_FRAME_FADETIME,
        HUD_LUP_FRAME_FADEOUTTIME);

      // Draw Leveled Text
      SetFont(HUD_LUP_TXT_LEVEL_FONT);
      HudMessage(
        s:txt_level;
        HUDMSG_FADEINOUT,
        txt_level_channel,
        txt_level_color,
        HUD_LUP_FRAME_X,
        txt_level_y,
        HUD_LUP_FRAME_DURATION,
        HUD_LUP_FRAME_FADETIME,
        HUD_LUP_FRAME_FADEOUTTIME);
    }

/* ****************************************************************
// REWARD VARIABLES (Configurable)
// ***************************************************************/
  // Level-up rewards:
  // PlayerClass # {Level # {Weapon ID #, Ammo ID #}}
  int djb_rewards[MAX_CLASSES][MAX_REWARDS][2] = {
    { // PlayerClass 0
      {1, 0}, // Level 1 -> Pistol, NONE
      {2, 2}, // Level 2 -> Shotgun, Shell
      {3, 1}, // Level 3 -> Chaingun, Clip
      {4, 2}, // Level 4 -> SuperShotgun, Shell
      {5, 1}, // Level 5 -> Minigun, Clip
      {6, 3}, // Level 6 -> RocketLauncher, RocketAmmo
      {7, 3}, // Level 7 -> GrenadeLauncher, RocketAmmo
	  {8, 4}, // Level 8 -> PlasmaRifle, Cell
      {9, 4}, // Level 9 -> Railgun, Cell
	  {10, 4}, // Level 10 -> BFG9000, Cell
	  {11, 4}, // Level 11 -> BFG10K, Cell 
	  {12, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4},
	  {13, 1},
	  {13, 2},
	  {13, 3},
	  {13, 4}
  }
};

  // Level-up reward: Weapon
  // classname, amount
  int djb_weapons[MAX_CLASSES][2][MAX_WEAPONS] = {
    { // PlayerClass 0
      {"Pistol", "PlasmaPistol" ,"Shotgun", "Chaingun2", "SuperShotgun", "Minigun", "RocketLauncher2", "GrenadeLauncher", "PlasmaRifle2", "Railgun", "BFG90002", "BFG10K", "ShotgunMKII"},
      {1,        1,              1,          1,           1,              1,         1,                 1,			       1,              1,         1,          1,        1}
    }
  };

  // Level-up reward: Ammo
  // classname, amount
  int djb_ammos[MAX_CLASSES][2][MAX_AMMOS] = {
    { // PlayerClass 0 
      {"Clip", "Shell", "RocketAmmo", "Cell"},
      {400,    100,     100,          600}
    }
  };
  
  
/* **************************************************************************
// CORE SCRIPTS (Don't change this)
// *************************************************************************/
  script 15 OPEN
  {
    int pc;

    // Weapons
    for (pc = 0; pc < MAX_CLASSES; pc++) {
      for (int w = 0; w < MAX_WEAPONS; w++) {
        str w_class = djb_weapons[pc][0][w];
        int w_amt = djb_weapons[pc][1][w];
        int w_num = w + 1;
        int wpc_num = pc + 1;
        SetCfgWeapon(wpc_num, w_num, w_class, w_amt);
      }
    }

    // Ammos
    for (pc = 0; pc < MAX_CLASSES; pc++) {
      for (int a = 0; a < MAX_AMMOS; a++) {
        str a_class = djb_ammos[pc][0][a];
        int a_amt = djb_ammos[pc][1][a];
        int a_num = a + 1;
        int apc_num = pc + 1;
        SetCfgAmmo(apc_num, a_num, a_class, a_amt);
      }
    }

    // Rewards
    for (pc = 0; pc < MAX_CLASSES; pc++) {
      for (int r = 0; r < MAX_REWARDS; r++) {
        int rw_num = djb_rewards[pc][r][0];
        int ra_num = djb_rewards[pc][r][1];
        int r_num = r + 1;
        int rpc_num = pc + 1;
        SetCfgReward(rpc_num, r_num, rw_num, ra_num);
      }
    }
	
	// XP
	int max_xp_incr_amt = GetCVAR("ad_max_xp_incr_amt");
	int max_xp_incr_amt_limit = GetCVAR("ad_max_xp_incr_amt_limit");
	if (max_xp_incr_amt < 100) {
	  max_xp_incr_amt = 100;
	  SetCVAR("ad_max_xp_incr_amt", max_xp_incr_amt);
	}
	if (max_xp_incr_amt_limit < 500) {
	  max_xp_incr_amt_limit = 500;
	  SetCVAR("ad_max_xp_incr_amt_limit", max_xp_incr_amt_limit);
	}
	SetCfgXpMaxIncrAmt(max_xp_incr_amt);
	SetCfgXpMaxIncrAmtLimit(max_xp_incr_amt_limit);

    // Done
    djb[1] = true;
  }



/* **************************************************************************
// CORE FUNCTIONS (Don't change these.)
// *************************************************************************/
  function void SetCfgWeapon(int pc_num, int w_num, str w_class, int w_amt) {
    // Weapon -> Player Class -> Weapon Number -> Weapon Name
    djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_WEAPON *xb*xc*xd+ pc_num *xc*xd+ w_num *xd+ IDX_CFG_ITEM_WEAPON_CLASS] = w_class;

    // Weapon -> Player Class -> Weapon Number -> Weapon Amount
    djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_WEAPON *xb*xc*xd+ pc_num *xc*xd+ w_num *xd+ IDX_CFG_ITEM_WEAPON_AMT] = w_amt;
  }

  function void SetCfgAmmo(int pc_num, int a_num, str a_class, int a_amt) {
    // Ammo -> Player Class -> Ammo Number -> Ammo Name
    djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_AMMO *xb*xc*xd+ pc_num *xc*xd+ a_num *xd+ IDX_CFG_ITEM_AMMO_CLASS] = a_class;

    // Ammo -> Player Class -> Ammo Number -> Ammo Amount
    djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_AMMO *xb*xc*xd+ pc_num *xc*xd+ a_num *xd+ IDX_CFG_ITEM_AMMO_AMT] = a_amt;
  }

  function void SetCfgReward(int pc_num, int r_num, int w_num, int a_num)
  {
    // Reward -> Player Class -> Reward Number -> Weapon Number
    // Reward -> Player Class -> Reward Number -> Ammo Number
    djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_REWARD *xb*xc*xd+ pc_num *xc*xd+ r_num *xd+ IDX_CFG_REWARD_WNUM] = w_num;
    djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_REWARD *xb*xc*xd+ pc_num *xc*xd+ r_num *xd+ IDX_CFG_REWARD_ANUM] = a_num;
  }
  
  
  function void SetCfgXpMaxIncrAmt(int amt) {
    djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_XP *xb*xc*xd+ 0 *xc*xd+ 0 *xd+ IDX_CFG_XP_MAX_INCR] = amt;
  }

  function void SetCfgXpMaxIncrAmtLimit(int amt) {
    djb[IDX_CFG_MAX_COUNT *xa*xb*xc*xd+ IDX_CFG_TYPE_XP *xb*xc*xd+ 0 *xc*xd+ 0 *xd+ IDX_CFG_XP_MAX_INCR_LIMIT] = amt;
  }
script 553(int HealType)
{//return maximum amount of health
	if(HealType<1)SetResultValue(CheckInventory("CurrentMaxHealth"));
	if(HealType==1)
	{//quartz flask
		SetResultValue(CheckInventory("CurrentMaxHealth")/6);
	}
}
script 250(int HealAmount)
{//this will heal a player until the healing ability of the item is gone, or their max health is reached
	SetActivatorToTarget(0);
	int Damage = CheckInventory("CurrentMaxHealth")-CheckInventory("Health");//amount of damage we've taken
	int OverFlow = CheckInventory("CurrentMaxHealth")-CheckInventory("Health")-HealAmount;//(the negative of)The amount by which healing by the desired amount will put us over our max health
	int H = HealAmount;
	//H = 10;
	if(OverFlow<0)H -= -OverFlow; //this will heal us by the desired amount, or up to our maximum
	//GiveInventory("NewHealth",H);
	//HealThing(H);
	SetActorProperty(0,APROP_Health,GetActorProperty(0,APROP_Health)+H);
}
  
function void HealingActor (int p_tid, int HealAmount)
{
	if(CheckActorInventory(p_tid,"CurrentMaxHealth"))
	{
		int Damage = CheckActorInventory(p_tid,"CurrentMaxHealth")-CheckActorInventory(p_tid,"Health");//amount of damage we've taken
		int OverFlow = CheckActorInventory(p_tid,"CurrentMaxHealth")-CheckActorInventory(p_tid,"Health")-HealAmount;//(the negative of)The amount by which healing by the desired amount will put us over our max health
		int H = HealAmount;
		if(OverFlow<0)H -= -OverFlow; //this will heal us by the desired amount, or up to our maximum
		If(GetActorProperty(p_tid,APROP_Health) > 0)
			SetActorProperty(p_tid,APROP_Health,GetActorProperty(p_tid,APROP_Health)+H);
	}
}
	
function void SetMaximumHealth (int p) 
{
      int p_tid = GetPlayerTIDByNum(p);
	SetActorProperty(p_tid,APROP_SpawnHEALTH,CheckInventory("CurrentMaxHealth"));
}

function void IncreaseMaximumHealth (int p) 
{    
	  int p_tid = GetPlayerTIDByNum(p);
		GiveInventory("CurrentMaxHealth",5);
		If(GetActorProperty(p_tid,APROP_Health) > 0)
			GiveInventory("CurrentMaxHealth",5);
	SetMaximumHealth(p);
}

script 123 RESPAWN
{
GiveInventory("TemporaryProtectionGiver",1);
SetMaximumHealth(0);
GiveInventory("BlueArmor",1);
SetActorProperty(0, APROP_Health, CheckInventory("Health"));
}

script 916 (void)
{
	SetResultValue(CheckInventory("CurrentMaxArmor"));
}

Script 787 ENTER Clientside
{
	SetMaximumHealth(0);
}

Script 788 RETURN Clientside
{
	SetMaximumHealth(0);
}
